/*!
 * dope 2.2.2
 * http://github.com/ryanve/dope
 * MIT License (c) Ryan Van Etten 
 */

/*jshint expr:true, sub:true, supernew:true, debug:true, node:true, boss:true, devel:true, evil:true, 
  laxcomma:true, eqnull:true, undef:true, unused:true, browser:true, jquery:true, maxerr:10 */

(function(root, name, make) {
  if (typeof module != 'undefined' && module.exports) module.exports = make();
  else root[name] = make();
}(this, 'dope', function() {

  // developers.google.com/closure/compiler/docs/api-tutorial3
  // developers.google.com/closure/compiler/docs/js-for-compiler

  var doc = document
    , xports = {}
    , effins = xports['fn'] = {}
    , owns = xports.hasOwnProperty
    , DMS = typeof DOMStringMap != 'undefined'
    , parseJSON = typeof JSON != 'undefined' && JSON.parse
    , queryMethod = 'querySelectorAll' 
    , QSA = !!doc[queryMethod] || !(queryMethod = 'getElementsByTagName')
    , queryEngine = function(s, root) {
        return s ? (root || doc)[queryMethod](s) : []; 
      }
    , camels = /([a-z])([A-Z])/g // lowercase next to uppercase
    , dashB4 = /-(.)/g // finds chars after hyphens
    , csvSsv = /\s*[\s\,]+\s*/ // splitter for comma *or* space-separated values
    , cleanAttr = /^[\[\s]+|\s+|[\]\s]+$/g // replace whitespace, trim [] brackets
    , cleanPre = /^[\[\s]?(data-)?|\s+|[\]\s]?$/g // replace whitespace, trim [] brackets, trim prefix
    , escDots = /\\*\./g // find periods w/ and w/o preceding backslashes
    , ssv = /\s+/
    , trimmer = /^\s+|\s+$/
    , trim = ''.trim ? function(s) {
        return null == s ? '' : s.trim(); 
      } : function(s) {
        return null == s ? '' : s.replace(trimmer, ''); 
      };
  
  /**
   * @return {string}
   */
  function camelHandler(all, letter) { 
    return letter.toUpperCase();
  }

  /**
   * Convert  'data-pulp-fiction' to 'pulpFiction'. Non-scalars return an
   * empty string. number|boolean coerces to string. (opposite: datatize())
   * @param {string|number|boolean|*} s
   * @return {string}
   */
  function camelize(s) {
    if (typeof s != 'string') return typeof s == 'number' || typeof s == 'boolean' ? '' + s : ''; 
    // Remove data- prefix and convert remaining dashed string to camelCase:
    return s.replace(cleanPre, '').replace(dashB4, camelHandler); // -a to A
  }

  /**
   * Convert 'pulpFiction' to 'data-pulp-fiction' OR 47 to 'data-47'
   * Invalid types return an empty string. (opposite: camelize())
   * @param {string|number|*} s
   * @return {string}
   */
  function datatize(s) {
    if (typeof s == 'string') s = s.replace(cleanPre, '$1').replace(camels, '$1-$2'); // aA to a-A
    else s = typeof s == 'number'  ? '' + s : '';
    return s ? ('data-' + s.toLowerCase()) : s;
  }

  /**
   * Convert a stringified primitive into its correct type.
   * @param {string|*} s
   */
  function parse(s) {
    var n; // undefined, or becomes number
    return typeof s != 'string' || !s ? s
      : 'false' === s ? false
      : 'true' === s ? true
      : 'null' === s ? null
      : 'undefined' === s || (n = (+s)) || 0 === n || 'NaN' === s ? n
      : s;
  }

  /**
   * @param {Object|Array|*} list
   * @param {Function} fn   
   * @param {(Object|*)=} scope
   * @param {boolean=} compact 
   * @return {Array}
   */
  function map(list, fn, scope, compact) {
    var l, i = 0, v, u = 0, ret = [];
    if (list == null) return ret;
    compact = true === compact;
    for (l = list.length; i < l;) {
      v = fn.call(scope, list[i], i++, list);
      if (v || !compact) ret[u++] = v;
    }
    return ret;
  }
  
  /** 
   * special-case DOM-node iterator optimized for internal use
   * @param {Object|Array} ob
   * @param {Function} fn
   * @param {*=} param
   */
  function eachNode(ob, fn, param) {
    for (var l = ob.length, i = 0; i < l; i++)
      ob[i] && ob[i].nodeType && fn(ob[i], param);
    return ob;
  }

  /**
   * internal-use function to iterate a node's attributes
   * @param {Object} el
   * @param {Function} fn
   * @param {(boolean|*)=} exp
   */
  function eachAttr(el, fn, exp) {
    var test, n, a, i, l;
    if (!el.attributes) return;
    test = typeof exp == 'boolean' ? /^data-/ : test;
    for (i = 0, l = el.attributes.length; i < l;) {
      if (a = el.attributes[i++]) {
        n = '' + a.name;
        test && test.test(n) !== exp || null == a.value || fn.call(el, a.value, n, a);
      }
    }
  }

  /**
   * Get object containing an element's data attrs.
   * @param {Element} el
   * @return {DOMStringMap|Object|undefined}
   */
  function getDataset(el) {
    var ob;
    if (!el || 1 !== el.nodeType) return;  // undefined
    if (ob = DMS && el.dataset) return ob; // native
    ob = {}; // Fallback plain object cannot mutate the dataset via reference.
    eachAttr(el, function(v, k) {
      ob[camelize(k)] = '' + v;
    }, true);
    return ob;
  }

  /**
   * @param {Element} el
   * @param {Object=} ob
   */
  function resetDataset(el, ob) {
    if (!el) return;
    var n, curr = el.dataset;
    if (curr && DMS) {
      if (curr === ob) return;
      for (n in curr) delete curr[n];
    }
    ob && dataset(el, ob);
  }
  
  /**
   * @param {Element} el
   * @param {Object} ob
   * @param {Function} fn
   */
  function setViaObject(el, ob, fn) {
    for (var n in ob) owns.call(ob, n) && fn(el, n, ob[n]);
  }
  
  /**
   * @param {Object|Array|Function} el
   * @param {(string|Object|*)=} k
   * @param {*=} v
   */  
  function attr(el, k, v) {
    el = el.nodeType ? el : el[0];
    if (!el || !el.setAttribute) return;
    k = typeof k == 'function' ? k.call(el) : k;
    if (!k) return;
    if (typeof k == 'object') {
      // SET-multi
      setViaObject(el, k, attr);
    } else {
      if (void 0 === v) {
        // GET
        k = el.getAttribute(k); // repurpose
        return null == k ? v : '' + k; // normalize
      }
      // SET
      v = typeof v == 'function' ? v.call(el) : v;
      v = '' + v; // normalize inputs
      el.setAttribute(k, v);
      return v; // the curr value
    }
  }
  
  /**
   * @param {Object|Array|Function} el
   * @param {(string|Object|*)=} k
   * @param {*=} v
   */  
  function dataset(el, k, v) {
    var exact, kFun = typeof k == 'function';
    el = el.nodeType ? el : el[0];
    if (!el || !el.setAttribute) return;
    if (void 0 === k && v === k) return getDataset(el);
    k = kFun ? k.call(el) : k;

    if (typeof k == 'object' && (kFun || !(exact = void 0 === v && datatize(k[0])))) {
      // SET-multi
      kFun && deletes(el);
      k && setViaObject(el, k, dataset);
    } else {
      k = exact || datatize(k);
      if (!k) return;
      if (void 0 === v) {
        // GET
        k = el.getAttribute(k); // repurpose
        return null == k ? v : exact ? parse(k) : '' + k; // normalize
      }
      // SET
      v = typeof v == 'function' ? v.call(el) : v;
      v = '' + v; // normalize inputs
      el.setAttribute(k, v);
      return v; // current value
    }
  }

  /**
   * @param {Element} el
   * @param {(Array|string|number)=} keys
   */
  function deletes(el, keys) {
    var k, i = 0;
    el = el.nodeType ? el : el[0];
    if (!el || !el.removeAttribute)
      return;
    if (void 0 === keys) {
      resetDataset(el); 
    } else {
      keys = typeof keys == 'string' ? keys.split(ssv) : [].concat(keys);
      while (i < keys.length) {
        k = datatize(keys[i++]);
        k && el.removeAttribute(k);
      }
    }
  }
  
  /**
   * @param {Element} el
   * @param {Array|string|number} keys
   */
  function removeAttr(el, keys) {
    var i = 0;
    el = el.nodeType ? el : el[0];
    if (el && el.removeAttribute) {
      for (keys = typeof keys == 'string' ? keys.split(ssv) : [].concat(keys); i < keys.length; i++) {
        keys[i] && el.removeAttribute(keys[i]);
      }
    }
  }

  /**
   * Convert list of attr names or data- keys into a selector.
   * @param {Array|string|number|*} list
   * @param {boolean=} prefix
   * @param {boolean=} join
   * @return {string|Array}
   */
  function toAttrSelector(list, prefix, join) {
    var l, s, i = 0, j = 0, emp = '', arr = [];
    prefix = true === prefix;
    list = typeof list == 'string' ? list.split(csvSsv) : typeof list == 'number' ? '' + list : list;
    for (l = list.length; i < l;) {
      s = list[i++];
      s = prefix ? datatize(s) : s.replace(cleanAttr, emp);
      s && (arr[j++] = s);
    }
    // Escape periods to allow atts like `[data-the.wh_o]`
    // @link api.jquery.com/category/selectors/
    // @link stackoverflow.com/q/13283699/770127
    return false === join ? arr : j ? '[' + arr.join('],[').replace(escDots, '\\\\.') + ']' : emp;
  }

  /**
   * Get elements matched by a data key.
   * @param {Array|string} list array or CSV or SSV data keys
   * @return {Array|*}
   */   
  xports['queryData'] = QSA ? function(list, root) {
    // Modern browsers, IE8+
    return false === root ? toAttrSelector(list, true, root) : queryEngine(toAttrSelector(list, true), root);
  } : function(list, root) {
    // == FALLBACK ==
    list = toAttrSelector(list, true, false);
    return false === root ? list : queryAttrFallback(list, root); 
  };
  
  /**
   * Get elements matched by an attribute name.
   * @param {Array|string} list array or CSV or SSV data keys
   * @return {Array|*}
   */   
  xports['queryAttr'] = QSA ? function(list, root) {
    // Modern browsers, IE8+
    return false === root ? toAttrSelector(list, root, root) : queryEngine(toAttrSelector(list), root);
  } : function(list, root) {
    // == FALLBACK ==
    list = toAttrSelector(list, false, false);
    return false === root ? list : queryAttrFallback(list, root); 
  };
  
  /**
   * @param {Array|string} list array of attribute names (w/o bracks)
   * @param {Object=} root
   */
  function queryAttrFallback(list, root) {
    var j, i, e, els, l = list.length, ret = [], u = 0;
    if (!l) return ret;
    els = queryEngine('*', root);
    for (j = 0; (e = els[j]); j++) {
      i = l; // reset i for each outer iteration
      while (i--) {// each attr name
        if (attr(e, list[i]) != null) {
          ret[u++] = e; // ghetto push
          break; // prevent pushing same elem twice
        }
      }
    }
    return ret;
  }
  
  // Expose remaining top-level methods:
  xports['map'] = map;
  xports['parse'] = parse;

  /**
   * @param {string|*} s
   * @since 2.1.0
   */
  xports['parseJSON'] = function(s) {
    s = parse(s);
    if (typeof s == 'string') {
      try {
        s = parseJSON(trim(s));
      } catch (e) {}
    }
    return s;
  };

  xports['trim'] = trim;
  xports['qsa'] = queryEngine;
  xports['attr'] = attr;
  xports['removeAttr'] = removeAttr;
  xports['dataset'] = dataset;
  xports['deletes'] = deletes;
  xports['camelize'] = camelize;
  xports['datatize'] = datatize;

  /**
   * @this {Object|Array}
   * @param {*=} k
   * @param {*=} v
   */
  effins['dataset'] = function(k, v) {
    var kMulti = typeof k == 'object' ? !(void 0 === v && datatize(k[0])) : typeof k == 'function';
    if (void 0 === v && !kMulti) return dataset(this[0], k); // GET
    return (k = kMulti ? k : datatize(k)) ? eachNode(this, function(e, x) {
      x = typeof v == 'function' ? v.call(e) : v;
      kMulti ? dataset(e, k, x) : e.setAttribute(k, '' + x); 
    }) : void 0 === v ? v : this;
  };

  /**
   * @this {Object|Array}
   * @param {*=} k
   * @param {*=} v
   */  
  effins['attr'] = function(k, v) {
    var kMulti = typeof k == 'object' || typeof k == 'function';
    if (void 0 === v && !kMulti) return attr(this[0], k); // GET
    return k ? eachNode(this, function(e, x) {
      x = typeof v == 'function' ? v.call(e) : v;
      kMulti ? attr(e, k, x) : e.setAttribute(k, '' + x); 
    }) : (void 0 === v ? v : this);
  };

  /**
   * Remove data- attrs for each element in a collection.
   * @this {Object|Array}
   * @param {Array|string}  keys  one or more SSV or CSV data attr keys or names
   */
  effins['deletes'] = function(keys) {
    if (void 0 === keys) return eachNode(this, resetDataset);
    keys = typeof keys == 'string' ? keys.split(ssv) : [].concat(keys);
    return eachNode(this, removeAttr, map(keys, datatize));
  };
  
  /**
   * Remove attrbutes for each element in a collection.
   * @this {Object|Array}
   * @param {Array|string}  keys  one or more SSV or CSV attr names
   */
  effins['removeAttr'] = function(keys) {
    return eachNode(this, removeAttr, keys);
  };

  return xports;
}));